module Aspect
  class Pointcut
    def |(other)
      OrPointcut.new(self, other)
    end

    def weave
      target_methods.each do |klass, method|
        advice = Module.new
        advice.send(:define_method, method) do |*args, &block|
          proceed = ->(*a, &b) { super(*a, &b) }
          yield(klass, method, args, block, proceed)
        end
        klass.send(:prepend, advice)
      end
    end
  end

  class ExecutionPointcut < Pointcut
    attr_reader :defined_class, :method_name

    def initialize(class:, method:)
      @defined_class = binding.local_variable_get(:class)
      @method_name = method
    end

    module Refinements
      refine Module do
        def match_class(klass)
          self == klass
        end
      end

      refine Regexp do
        def match_class(klass)
          self =~ klass.name
        end
      end
    end

    using Refinements

    def target_methods
      ObjectSpace.each_object(Class).select { |klass|
        @defined_class.match_class(klass)
      }.flat_map { |klass|
        klass.instance_methods(false).select { |method|
          @method_name === method
        }.map { |method|
          [klass, method]
        }
      }
    end
  end

  class OrPointcut < Pointcut
    def initialize(left, right)
      @left = left
      @right = right
    end

    def target_methods
      @left.target_methods | @right.target_methods
    end
  end

  def execution(*args)
    ExecutionPointcut.new(*args)
  end

  class JoinPoint
    attr_reader :defined_class, :method, :args, :block

    def initialize(defined_class, method, args, block)
      @defined_class = defined_class
      @method = method
      @args = args
      @block = block
    end

    def signature
      "#{@defined_class.inspect}##{@method}"
    end
  end

  class ProceedingJoinPoint < JoinPoint
    def initialize(*args, proceed)
      super(*args)
      @proceed = proceed
    end

    def proceed(args = @args, block = @block)
      @proceed.call(*args, &block)
    end
  end

  def before(pointcut)
    pointcut.weave do |klass, method, args, block, proceed|
      join_point = JoinPoint.new(klass, method, args, block)
      yield(join_point)
      proceed.call(*args, &block)
    end
  end

  def after(pointcut)
    pointcut.weave do |klass, method, args, block, proceed|
      begin
        proceed.call(*args, &block)
      ensure
        join_point = JoinPoint.new(klass, method, args, block)
        yield(join_point)
      end
    end
  end

  def around(pointcut)
    pointcut.weave do |klass, method, args, block, proceed|
      join_point = ProceedingJoinPoint.new(klass, method, args,
                                           block, proceed)
      yield(join_point)
    end
  end
end
